/******************************************************************
MIT License http://www.opensource.org/licenses/mit-license.php
Author Qiming Zhao <chemzqm@gmail> (https://github.com/chemzqm)
*******************************************************************/
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const wxmlScanner_1 = require("./wxmlScanner");
const arrays_1 = require("../utils/arrays");
const wxmlTags_1 = require("./wxmlTags");
const wxmlLanguageTypes_1 = require("../wxmlLanguageTypes");
function isValidTagName(tag) {
    return /^[a-z]([a-z]|\d|-|_)*$/.test(tag);
}
function isValidAttribute(attr) {
    return /^[a-zA-Z](:|\.|-|_|\w)*$/.test(attr);
}
class Node {
    constructor(start, end, children, parent) {
        this.start = start;
        this.end = end;
        this.children = children;
        this.parent = parent;
        this.closed = false;
        this.textContents = [];
    }
    get attributeNames() {
        return this.attributes ? Object.keys(this.attributes).map(s => s.replace(/\|\d+$/, '')) : [];
    }
    isSameTag(tagInLowerCase) {
        if (!this.tag || !tagInLowerCase) {
            return false;
        }
        return (this.tag.length === tagInLowerCase.length &&
            this.tag.toLowerCase() === tagInLowerCase);
    }
    getAttributeAtOffset(offset) {
        if (!this.attributes)
            return null;
        for (let attr of Object.keys(this.attributes)) {
            let parts = attr.split('|');
            if (parts.length < 2)
                continue;
            let start = parseInt(parts[1], 10);
            let end = start + parts[0].length;
            if (offset >= start && offset <= end) {
                return attr;
            }
        }
        return null;
    }
    getAttributeOffsets(name) {
        if (!this.attributes)
            return [];
        let res = [];
        for (let attr of Object.keys(this.attributes)) {
            let parts = attr.split('|');
            if (parts[0] == name) {
                res.push(parseInt(parts[1], 10));
            }
        }
        return res;
    }
    getDuplicatedAttrs() {
        let names = this.attributeNames;
        let res = [];
        for (let name of names) {
            if (res.indexOf(name) === -1 && names.indexOf(name) !== names.lastIndexOf(name)) {
                res.push(name);
            }
        }
        return res;
    }
    getValueOffset(attributeName) {
        if (!this.attributes)
            return null;
        for (let attr of Object.keys(this.attributes)) {
            let parts = attr.split('|');
            if (attributeName === parts[0]) {
                let value = this.attributes[attr];
                if (value || value == '') {
                    return parseInt(parts[1], 10) + attributeName.length + 2;
                }
            }
        }
        return null;
    }
    getAttributeValue(attributeName) {
        if (!this.attributes)
            return null;
        for (let attr of Object.keys(this.attributes)) {
            let parts = attr.split('|');
            if (attributeName === parts[0]) {
                let value = this.attributes[attr];
                return value;
            }
        }
        return null;
    }
    get isEmpty() {
        return this.children.length === 0 && this.textContents.length === 0;
    }
    get firstChild() {
        return this.children[0];
    }
    get lastChild() {
        return this.children.length
            ? this.children[this.children.length - 1]
            : void 0;
    }
    findNodeBefore(offset) {
        let idx = arrays_1.findFirst(this.children, c => offset <= c.start) - 1;
        if (idx >= 0) {
            let child = this.children[idx];
            if (offset > child.start) {
                if (offset < child.end) {
                    return child.findNodeBefore(offset);
                }
                let lastChild = child.lastChild;
                if (lastChild && lastChild.end === child.end) {
                    return child.findNodeBefore(offset);
                }
                return child;
            }
        }
        return this;
    }
    findNodeAt(offset) {
        let idx = arrays_1.findFirst(this.children, c => offset <= c.start) - 1;
        if (idx >= 0) {
            let child = this.children[idx];
            if (offset > child.start && offset <= child.end) {
                return child.findNodeAt(offset);
            }
        }
        return this;
    }
}
exports.Node = Node;
function parse(text) {
    let scanner = wxmlScanner_1.createScanner(text);
    let wxmlDocument = new Node(0, text.length, [], void 0);
    let curr = wxmlDocument;
    let endTagStart = -1;
    let pendingAttribute = null;
    let token = scanner.scan();
    let errors = [];
    let tagStack = [];
    let currOffset = -1;
    while (token !== wxmlLanguageTypes_1.TokenType.EOS) {
        let errorMsg = scanner.getTokenError();
        if (errorMsg) {
            errors.push({
                message: errorMsg,
                offsetStart: scanner.getTokenOffset(),
                offeetEnd: scanner.getTokenEnd()
            });
        }
        // let text = scanner.getTokenType()
        // console.log(text)
        switch (token) {
            case wxmlLanguageTypes_1.TokenType.StartTagOpen:
                let child = new Node(scanner.getTokenOffset(), text.length, [], curr);
                curr.children.push(child);
                curr = child;
                break;
            case wxmlLanguageTypes_1.TokenType.StartTag:
                curr.tag = scanner.getTokenText();
                tagStack.push(`${curr.tag}|${scanner.getTokenOffset()}|${scanner.getTokenEnd()}`);
                if (!isValidTagName(curr.tag)) {
                    errors.push({
                        message: `Only [a-z] _ - allowed for tag name`,
                        offsetStart: scanner.getTokenOffset(),
                        offeetEnd: scanner.getTokenEnd()
                    });
                }
                break;
            case wxmlLanguageTypes_1.TokenType.StartTagClose:
                curr.end = scanner.getTokenEnd(); // might be later set to end tag position
                if (curr.tag && wxmlTags_1.isEmptyElement(curr.tag) && curr.parent) {
                    curr.closed = true;
                    curr = curr.parent;
                }
                break;
            case wxmlLanguageTypes_1.TokenType.EndTagOpen:
                endTagStart = scanner.getTokenOffset();
                break;
            case wxmlLanguageTypes_1.TokenType.EndTag:
                let closeTag = scanner.getTokenText().toLowerCase();
                while (!curr.isSameTag(closeTag) && curr.parent) {
                    curr.end = endTagStart;
                    curr.closed = false;
                    curr = curr.parent;
                }
                if (curr !== wxmlDocument) {
                    curr.closed = true;
                    curr.endTagStart = endTagStart;
                }
                let last = tagStack.pop();
                if (!last) {
                    errors.push({
                        message: `Start tag not found for <${closeTag}>`,
                        offsetStart: scanner.getTokenOffset(),
                        offeetEnd: scanner.getTokenEnd()
                    });
                }
                else {
                    let parts = last.toLowerCase().split('|');
                    if (closeTag !== parts[0]) {
                        errors.push({
                            message: `Close tag not found for <${parts[0]}>`,
                            offsetStart: parseInt(parts[1], 10),
                            offeetEnd: parseInt(parts[2], 10),
                        });
                    }
                }
                break;
            case wxmlLanguageTypes_1.TokenType.StartTagSelfClose:
                if (curr.parent) {
                    curr.closed = true;
                    curr.end = scanner.getTokenEnd();
                    curr = curr.parent;
                }
                tagStack.pop();
                break;
            case wxmlLanguageTypes_1.TokenType.EndTagClose:
                if (curr.parent) {
                    curr.end = scanner.getTokenEnd();
                    curr = curr.parent;
                }
                break;
            case wxmlLanguageTypes_1.TokenType.AttributeName: {
                pendingAttribute = scanner.getTokenText();
                if (!isValidAttribute(pendingAttribute)) {
                    errors.push({
                        message: `Invalid attribute ${pendingAttribute}`,
                        offsetStart: scanner.getTokenOffset(),
                        offeetEnd: scanner.getTokenEnd() - 1
                    });
                }
                let attributes = curr.attributes;
                if (!attributes) {
                    curr.attributes = attributes = {};
                }
                currOffset = scanner.getTokenOffset();
                if (pendingAttribute) {
                    attributes[pendingAttribute + `|${currOffset}`] = null; // Support valueless attributes such as 'checked'
                }
                break;
            }
            case wxmlLanguageTypes_1.TokenType.AttributeValue: {
                let value = scanner.getTokenText();
                if (!/^".*"$/.test(value)) {
                    errors.push({
                        message: `Double quote required for value`,
                        offsetStart: scanner.getTokenOffset(),
                        offeetEnd: scanner.getTokenEnd() - 1
                    });
                }
                let attributes = curr.attributes;
                if (attributes && pendingAttribute) {
                    attributes[pendingAttribute + `|${currOffset}`] = value.replace(/^"/, '').replace(/"$/, '');
                    pendingAttribute = null;
                }
                break;
            }
            case wxmlLanguageTypes_1.TokenType.Content: {
                let text = scanner.getTokenText().trim();
                if (text) {
                    curr.textContents.push(text);
                }
                break;
            }
        }
        token = scanner.scan();
    }
    while (curr.parent) {
        curr.end = text.length;
        curr.closed = false;
        curr = curr.parent;
    }
    if (tagStack.length) {
        for (let item of tagStack) {
            let parts = item.toLowerCase().split('|');
            errors.push({
                message: `Close tag not found for <${parts[0]}>`,
                offsetStart: parseInt(parts[1], 10),
                offeetEnd: parseInt(parts[2], 10) - 1,
            });
        }
    }
    return {
        roots: wxmlDocument.children,
        findNodeBefore: wxmlDocument.findNodeBefore.bind(wxmlDocument),
        findNodeAt: wxmlDocument.findNodeAt.bind(wxmlDocument),
        errors,
    };
}
exports.parse = parse;
//# sourceMappingURL=wxmlParser.js.map