"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/******************************************************************
MIT License http://www.opensource.org/licenses/mit-license.php
Author Qiming Zhao <chemzqm@gmail> (https://github.com/chemzqm)
*******************************************************************/
const arrays = require("../utils/arrays");
const strings = require("../utils/strings");
const fs = require("fs");
const path = require("path");
exports.EMPTY_ELEMENTS = [
    'audio',
    'video',
    'camera',
    'checkbox',
    'cover-image',
    'icon',
    'image',
    'input',
    'live-player',
    'live-pusher',
    'map',
    'open-data',
    'progress',
    'radio',
    'slider',
    'switch',
];
function isEmptyElement(e) {
    return (!!e &&
        arrays.binarySearch(exports.EMPTY_ELEMENTS, e.toLowerCase(), (s1, s2) => s1.localeCompare(s2)) >= 0);
}
exports.isEmptyElement = isEmptyElement;
function getSubAttribute(spec) {
    let attrs = spec.attrs;
    if (!attrs)
        return null;
    for (let attr of attrs) {
        if (attr.hasOwnProperty('subAttrs')) {
            return attr;
        }
    }
    return null;
}
class WXMLTagSpecification {
    constructor(config) {
        this.attrs = [];
        for (let key of Object.keys(config)) {
            this[key] = config[key];
        }
    }
}
exports.WXMLTagSpecification = WXMLTagSpecification;
function isSubAttrTag(tag) {
    for (let name of Object.keys(exports.WXML_TAGS)) {
        if (name === tag) {
            return exports.WXML_TAGS[name].subAttr === true;
        }
    }
    return false;
}
exports.isSubAttrTag = isSubAttrTag;
exports.WXML_TAGS = (() => {
    let componentConfig = fs.readFileSync(path.resolve(__dirname, '../../components.json'), 'utf8');
    let componentList = JSON.parse(componentConfig);
    let TAGS = {};
    for (let item of componentList) {
        TAGS[item.name] = new WXMLTagSpecification(item);
        let subAttr = getSubAttribute(item);
        if (subAttr) {
            TAGS[item.name].subAttr = true;
            let attrs = subAttr.subAttrs;
            for (let attr of attrs) {
                let name = `$${item.name} ${attr.equal}`;
                let spec = {
                    name,
                    attrs: attr.attrs,
                };
                TAGS[name] = new WXMLTagSpecification(spec);
            }
        }
        if (item.name === 'input' || item.name === 'textarea') {
            let spec = TAGS[item.name];
            spec.attrs.push({
                name: 'name',
                type: {
                    name: 'string',
                },
                desc: ['field name']
            });
        }
    }
    TAGS.wxs = new WXMLTagSpecification({
        name: 'wxs',
        desc: ['WeiXin Script'],
        docLink: 'https://developers.weixin.qq.com/miniprogram/dev/framework/view/wxs/index.html',
        tips: ['WXS（WeiXin Script）是小程序的一套脚本语言，结合 WXML，可以构建出页面的结构。'],
        attrs: [{
                name: 'module',
                type: {
                    name: 'string',
                },
                desc: [
                    'global variable name'
                ],
            }]
    });
    return TAGS;
})();
function getWXMLTagProvider() {
    let globalAttributes = [
        'id:string',
        'class:string',
        'style:string',
        'hidden:boolean'
    ];
    let eventNames = ['tap', 'touchstart', 'touchmove', 'touchcancel', 'touchend',
        'longpress', 'longtap', 'transitionend', 'animationstart',
        'animationiteration', 'animationend', 'touchforcechange'];
    let eventHandlers = [];
    let logicAttributes = ['wx:if', 'wx:elif', 'wx:else'];
    let iterateAttributes = ['wx:for', 'wx:item', 'wx:index', 'wx:key', 'wx:for-item', 'wx:for-index', 'wx:for-items'];
    for (let name of eventNames) {
        eventHandlers.push(`bind:${name}`, `catch:${name}`, `catch${name}`, `bind${name}`);
    }
    let valueSets = {
        boolean: ['true', 'false']
    };
    return {
        getId: () => 'wxml',
        isApplicable: () => true,
        collectTags: (collector) => collectTagsDefault(collector, exports.WXML_TAGS),
        collectAttributes: (tag, collector) => {
            collectAttributesDefault(tag, collector, exports.WXML_TAGS, globalAttributes);
            for (let handler of eventHandlers) {
                collector(handler, 'event');
            }
            for (let attr of logicAttributes) {
                collector(attr, 'logic');
            }
            for (let attr of iterateAttributes) {
                collector(attr, 'iterate');
            }
        },
        collectValues: (tag, attribute, collector) => collectValuesDefault(tag, attribute, collector, exports.WXML_TAGS, globalAttributes, valueSets),
    };
}
exports.getWXMLTagProvider = getWXMLTagProvider;
function collectTagsDefault(collector, tagSet) {
    for (let tag of Object.keys(tagSet)) {
        // tag of sub attr
        if (tag[0] !== '$') {
            collector(tag, tagSet[tag]);
        }
    }
}
function collectAttributesDefault(tagName, collector, tagSet, globalAttributes) {
    if (tagName) {
        let tag = tagSet[tagName];
        if (tag) {
            for (let attr of tag.attrs) {
                collector(attr.name, attr.type.name, attr);
            }
        }
    }
    for (let attr of globalAttributes) {
        let segments = attr.split(':');
        collector(segments[0], segments[1]);
    }
}
function collectValuesDefault(tagName, attribute, collector, tagSet, globalAttributes, valueSets) {
    let prefix = attribute + ':';
    for (let attr of globalAttributes) {
        if (attr.length > prefix.length && strings.startsWith(attr, prefix)) {
            let typeInfo = attr.substr(prefix.length);
            let values = valueSets[typeInfo];
            if (values) {
                for (let value of values) {
                    collector(value);
                }
            }
            return;
        }
    }
    if (tagName) {
        let tag = tagSet[tagName];
        if (tag) {
            let attr = tag.attrs.find(o => o.name === attribute);
            if (attr) {
                if (attr.type.name === 'boolean') {
                    let values = valueSets.boolean;
                    for (let value of values) {
                        collector(value);
                    }
                }
                if (Array.isArray(attr.enum)) {
                    for (let item of attr.enum) {
                        collector(item.value, item);
                    }
                }
            }
        }
    }
}
//# sourceMappingURL=wxmlTags.js.map