"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_languageserver_protocol_1 = require("vscode-languageserver-protocol");
const manager_1 = require("./manager");
function createServer(connection) {
    const documents = new vscode_languageserver_1.TextDocuments();
    const manager = manager_1.getManager();
    let shouldShutdown = false;
    connection.onShutdown(() => {
        shouldShutdown = true;
    });
    process.on('exit', code => {
        if (code !== 0) {
            connection.sendNotification('window/showMessage', {
                type: vscode_languageserver_protocol_1.MessageType.Error,
                message: `wxml language server abnormal exit with code ${code}`
            });
        }
    });
    connection.onExit(() => {
        connection.sendNotification('window/showMessage', {
            type: vscode_languageserver_protocol_1.MessageType.Info,
            message: 'wxml service exited'
        });
        if (shouldShutdown) {
            process.exit();
        }
        else {
            process.exit(1);
        }
    });
    connection.onInitialize((params) => {
        let snippetSupport = false;
        const initializationOptions = params.initializationOptions;
        if (initializationOptions) {
            const { textDocument } = params.capabilities;
            if (textDocument
                && textDocument.completion
                && textDocument.completion.completionItem
                && textDocument.completion.completionItem.snippetSupport) {
                snippetSupport = true;
            }
        }
        manager.setConfig({
            complete: {
                completeEvent: true,
                useSnippet: snippetSupport
            }
        });
        documents.onDidClose(e => {
            manager.removeDocument(e.document.uri);
        });
        const capabilities = {
            // Tell the client that the server works in FULL text document sync mode
            textDocumentSync: documents.syncKind,
            completionProvider: {
                triggerCharacters: ['<'],
                resolveProvider: false,
            },
            hoverProvider: true,
        };
        return { capabilities };
    });
    // The settings have changed. Is send on server activation as well.
    connection.onDidChangeConfiguration((change) => {
        let config = change.settings;
        manager.setConfig(config);
    });
    const pendingValidationRequests = {};
    const validationDelayMs = 100;
    documents.onDidOpen((change) => {
        triggerValidation(change.document);
    });
    // When the text document first opened or when its content has changed.
    documents.onDidChangeContent((change) => {
        triggerValidation(change.document);
    });
    // A document has closed: clear all diagnostics
    documents.onDidClose((event) => {
        cleanPendingValidation(event.document);
        connection.sendDiagnostics({ uri: event.document.uri, diagnostics: [] });
    });
    function cleanPendingValidation(textDocument) {
        const request = pendingValidationRequests[textDocument.uri];
        if (request) {
            clearTimeout(request);
            delete pendingValidationRequests[textDocument.uri];
        }
    }
    function triggerValidation(textDocument) {
        cleanPendingValidation(textDocument);
        pendingValidationRequests[textDocument.uri] = setTimeout(() => {
            delete pendingValidationRequests[textDocument.uri];
            validateTextDocument(textDocument);
        }, validationDelayMs);
    }
    function validateTextDocument(textDocument) {
        const diagnostics = manager.doDiagnostic(textDocument);
        connection.sendDiagnostics({ uri: textDocument.uri, diagnostics });
    }
    connection.onCompletion(textDocumentPosition => {
        const document = documents.get(textDocumentPosition.textDocument.uri);
        if (!document)
            return null;
        return manager.doComplete(document, textDocumentPosition.position);
    });
    connection.onCompletionResolve(item => {
        return item;
    });
    connection.onHover(textDocumentPosition => {
        const document = documents.get(textDocumentPosition.textDocument.uri);
        if (!document)
            return null;
        return manager.doHover(document, textDocumentPosition.position);
    });
    return {
        listen() {
            documents.listen(connection);
            connection.listen();
        }
    };
}
exports.createServer = createServer;
//# sourceMappingURL=server.js.map