/******************************************************************
MIT License http://www.opensource.org/licenses/mit-license.php
Author Qiming Zhao <chemzqm@gmail> (https://github.com/chemzqm)
 *******************************************************************/
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const wxmlScanner_1 = require("../parser/wxmlScanner");
const wxmlTags_1 = require("../parser/wxmlTags");
const wxmlTags_2 = require("../parser/wxmlTags");
const wxmlLanguageTypes_1 = require("../wxmlLanguageTypes");
const htmlEntities_1 = require("../parser/htmlEntities");
const strings_1 = require("../utils/strings");
function doComplete(document, position, wxmlDocument, config) {
    let result = {
        isIncomplete: false,
        items: [],
    };
    let provider = wxmlTags_2.getWXMLTagProvider();
    let text = document.getText();
    let offset = document.offsetAt(position);
    let node = wxmlDocument.findNodeBefore(offset);
    if (!node)
        return result;
    let scanner = wxmlScanner_1.createScanner(text, node.start);
    let currentTag = '';
    let currentAttributeName;
    function getReplaceRange(replaceStart, replaceEnd = offset) {
        if (replaceStart > offset) {
            replaceStart = offset;
        }
        return {
            start: document.positionAt(replaceStart),
            end: document.positionAt(replaceEnd),
        };
    }
    function collectOpenTagSuggestions(afterOpenBracket, tagNameEnd) {
        let range = getReplaceRange(afterOpenBracket, tagNameEnd);
        provider.collectTags((tag, info) => {
            result.items.push({
                label: tag,
                kind: vscode_languageserver_types_1.CompletionItemKind.Property,
                documentation: info && info.desc ? info.desc.join('\n') : '',
                textEdit: vscode_languageserver_types_1.TextEdit.replace(range, tag),
                insertTextFormat: vscode_languageserver_types_1.InsertTextFormat.PlainText,
            });
        });
        return result;
    }
    function getLineIndent(offset) {
        let start = offset;
        while (start > 0) {
            let ch = text.charAt(start - 1);
            if ('\n\r'.indexOf(ch) >= 0) {
                return text.substring(start, offset);
            }
            if (!isWhiteSpace(ch)) {
                return null;
            }
            start--;
        }
        return text.substring(0, offset);
    }
    function collectCloseTagSuggestions(afterOpenBracket, inOpenTag, tagNameEnd = offset) {
        let range = getReplaceRange(afterOpenBracket, tagNameEnd);
        let closeTag = isFollowedBy(text, tagNameEnd, wxmlLanguageTypes_1.ScannerState.WithinEndTag, wxmlLanguageTypes_1.TokenType.EndTagClose) ? '' : '>';
        let curr = node;
        if (inOpenTag) {
            curr = curr.parent; // don't suggest the own tag, it's not yet open
        }
        while (curr) {
            let tag = curr.tag;
            if (tag &&
                (!curr.closed || (curr.endTagStart && curr.endTagStart > offset))) {
                let item = {
                    label: '/' + tag,
                    kind: vscode_languageserver_types_1.CompletionItemKind.Property,
                    filterText: '/' + tag + closeTag,
                    textEdit: vscode_languageserver_types_1.TextEdit.replace(range, '/' + tag + closeTag),
                    insertTextFormat: vscode_languageserver_types_1.InsertTextFormat.PlainText,
                };
                let startIndent = getLineIndent(curr.start);
                let endIndent = getLineIndent(afterOpenBracket - 1);
                if (startIndent !== null &&
                    endIndent !== null &&
                    startIndent !== endIndent) {
                    let insertText = startIndent + '</' + tag + closeTag;
                    item.textEdit = vscode_languageserver_types_1.TextEdit.replace(getReplaceRange(afterOpenBracket - 1 - endIndent.length), insertText);
                    item.filterText = endIndent + '</' + tag + closeTag;
                }
                result.items.push(item);
                return result;
            }
            curr = curr.parent;
        }
        if (inOpenTag) {
            return result;
        }
        provider.collectTags((tag, info) => {
            result.items.push({
                label: '/' + tag,
                kind: vscode_languageserver_types_1.CompletionItemKind.Property,
                documentation: info && info.desc ? info.desc.join('\n') : '',
                filterText: '/' + tag + closeTag,
                textEdit: vscode_languageserver_types_1.TextEdit.replace(range, '/' + tag + closeTag),
                insertTextFormat: vscode_languageserver_types_1.InsertTextFormat.PlainText,
            });
        });
        return result;
    }
    function collectAutoCloseTagSuggestion(tagCloseEnd, tag) {
        if (!wxmlTags_1.isEmptyElement(tag)) {
            let pos = document.positionAt(tagCloseEnd);
            result.items.push({
                label: '</' + tag + '>',
                kind: vscode_languageserver_types_1.CompletionItemKind.Property,
                filterText: '</' + tag + '>',
                textEdit: vscode_languageserver_types_1.TextEdit.insert(pos, '$0</' + tag + '>'),
                insertTextFormat: vscode_languageserver_types_1.InsertTextFormat.Snippet,
            });
        }
        return result;
    }
    function collectTagSuggestions(tagStart, tagEnd) {
        collectOpenTagSuggestions(tagStart, tagEnd);
        collectCloseTagSuggestions(tagStart, true, tagEnd);
        return result;
    }
    function collectAttributeNameSuggestions(tag, nameStart, nameEnd = offset) {
        if (!wxmlTags_2.isSubAttrTag(tag)) {
            let replaceEnd = offset;
            while (replaceEnd < nameEnd && text[replaceEnd] !== '<') {
                // < is a valid attribute name character, but we rather assume the attribute name ends. See #23236.
                replaceEnd++;
            }
            let isSnippet = !isFollowedBy(text, nameEnd, wxmlLanguageTypes_1.ScannerState.AfterAttributeName, wxmlLanguageTypes_1.TokenType.DelimiterAssign) && config.useSnippet;
            let range = getReplaceRange(nameStart, replaceEnd);
            let value = isSnippet ? '="$1"' : '';
            let seenAttributes = Object.create(null);
            for (let attr of node.attributeNames) {
                seenAttributes[attr] = true;
            }
            provider.collectAttributes(tag, (attribute, type, info) => {
                if (seenAttributes[attribute])
                    return;
                seenAttributes[attribute] = true;
                if (/^(bind|catch):/.test(attribute)) {
                    seenAttributes[attribute.replace(':', '')] = true;
                    if (config.completeEvent === false)
                        return;
                }
                if (/^\$/.test(tag) && /:/.test(attribute))
                    return;
                let codeSnippet = attribute;
                let useSnippet = isSnippet && type !== 'boolean';
                if (type !== 'boolean' && isSnippet) {
                    codeSnippet = codeSnippet + value;
                }
                let command;
                if (useSnippet) {
                    command = {
                        title: 'Suggest',
                        command: 'editor.action.triggerSuggest'
                    };
                }
                result.items.push({
                    label: attribute,
                    documentation: info && info.desc ? info.desc.join('\n') : '',
                    kind: vscode_languageserver_types_1.CompletionItemKind.Property,
                    textEdit: vscode_languageserver_types_1.TextEdit.replace(range, codeSnippet),
                    insertTextFormat: useSnippet ? vscode_languageserver_types_1.InsertTextFormat.Snippet : vscode_languageserver_types_1.InsertTextFormat.PlainText,
                    command
                });
            });
            return result;
        }
        else {
            let mode = node.getAttributeValue('mode');
            let name = mode ? `$${tag} ${mode}` : tag;
            return collectAttributeNameSuggestions(name, nameStart, nameEnd);
        }
    }
    function collectAttributeValueSuggestions(valueStart, valueEnd = offset) {
        let range;
        let addQuotes;
        if (offset > valueStart && offset <= valueEnd && isQuote(text[valueStart])) {
            // inside quoted attribute
            let valueContentStart = valueStart + 1;
            let valueContentEnd = valueEnd;
            // valueEnd points to the char after quote, which encloses the replace range
            if (valueEnd > valueStart && text[valueEnd - 1] === text[valueStart]) {
                valueContentEnd--;
            }
            let wsBefore = getWordStart(text, offset, valueContentStart);
            let wsAfter = getWordEnd(text, offset, valueContentEnd);
            range = getReplaceRange(wsBefore, wsAfter);
            addQuotes = false;
        }
        else {
            range = getReplaceRange(valueStart, valueEnd);
            addQuotes = true;
        }
        let tag = currentTag.toLowerCase();
        let attribute = currentAttributeName.toLowerCase();
        provider.collectValues(tag, attribute, (value, info) => {
            let insertText = addQuotes ? '"' + value + '"' : value;
            result.items.push({
                label: value,
                filterText: insertText,
                kind: vscode_languageserver_types_1.CompletionItemKind.Enum,
                documentation: info && info.desc ? info.desc : '',
                textEdit: vscode_languageserver_types_1.TextEdit.replace(range, insertText),
                insertTextFormat: vscode_languageserver_types_1.InsertTextFormat.PlainText,
            });
        });
        return result;
    }
    function scanNextForEndPos(nextToken) {
        if (offset === scanner.getTokenEnd()) {
            token = scanner.scan();
            if (token === nextToken && scanner.getTokenOffset() === offset) {
                return scanner.getTokenEnd();
            }
        }
        return offset;
    }
    function collectCharacterEntityProposals() {
        // character entities
        let k = offset - 1;
        let characterStart = position.character;
        while (k >= 0 && strings_1.isLetterOrDigit(text, k)) {
            k--;
            characterStart--;
        }
        if (k >= 0 && text[k] === '&') {
            let range = vscode_languageserver_types_1.Range.create(vscode_languageserver_types_1.Position.create(position.line, characterStart - 1), position);
            for (let entity in htmlEntities_1.entities) {
                if (strings_1.endsWith(entity, ';')) {
                    const label = '&' + entity;
                    result.items.push({
                        label,
                        kind: vscode_languageserver_types_1.CompletionItemKind.Keyword,
                        documentation: `Character entity representing '${htmlEntities_1.entities[entity]}'`,
                        textEdit: vscode_languageserver_types_1.TextEdit.replace(range, label),
                        insertTextFormat: vscode_languageserver_types_1.InsertTextFormat.PlainText,
                    });
                }
            }
        }
        return result;
    }
    let token = scanner.scan();
    while (token !== wxmlLanguageTypes_1.TokenType.EOS && scanner.getTokenOffset() <= offset) {
        switch (token) {
            case wxmlLanguageTypes_1.TokenType.StartTagOpen:
                if (scanner.getTokenEnd() === offset) {
                    let endPos = scanNextForEndPos(wxmlLanguageTypes_1.TokenType.StartTag);
                    return collectTagSuggestions(offset, endPos);
                }
                break;
            case wxmlLanguageTypes_1.TokenType.StartTag:
                if (scanner.getTokenOffset() <= offset &&
                    offset <= scanner.getTokenEnd()) {
                    return collectOpenTagSuggestions(scanner.getTokenOffset(), scanner.getTokenEnd());
                }
                currentTag = scanner.getTokenText();
                break;
            case wxmlLanguageTypes_1.TokenType.AttributeName:
                if (scanner.getTokenOffset() <= offset &&
                    offset <= scanner.getTokenEnd()) {
                    return collectAttributeNameSuggestions(currentTag.toLowerCase(), scanner.getTokenOffset(), scanner.getTokenEnd());
                }
                currentAttributeName = scanner.getTokenText();
                break;
            case wxmlLanguageTypes_1.TokenType.DelimiterAssign:
                if (scanner.getTokenEnd() === offset) {
                    let endPos = scanNextForEndPos(wxmlLanguageTypes_1.TokenType.AttributeValue);
                    return collectAttributeValueSuggestions(offset, endPos);
                }
                break;
            case wxmlLanguageTypes_1.TokenType.AttributeValue:
                if (scanner.getTokenOffset() <= offset &&
                    offset <= scanner.getTokenEnd()) {
                    return collectAttributeValueSuggestions(scanner.getTokenOffset(), scanner.getTokenEnd());
                }
                break;
            case wxmlLanguageTypes_1.TokenType.Whitespace:
                if (offset <= scanner.getTokenEnd()) {
                    switch (scanner.getScannerState()) {
                        case wxmlLanguageTypes_1.ScannerState.AfterOpeningStartTag:
                            let startPos = scanner.getTokenOffset();
                            let endTagPos = scanNextForEndPos(wxmlLanguageTypes_1.TokenType.StartTag);
                            return collectTagSuggestions(startPos, endTagPos);
                        case wxmlLanguageTypes_1.ScannerState.WithinTag:
                        case wxmlLanguageTypes_1.ScannerState.AfterAttributeName:
                            return collectAttributeNameSuggestions(currentTag.toLowerCase(), scanner.getTokenEnd());
                        case wxmlLanguageTypes_1.ScannerState.BeforeAttributeValue:
                            return collectAttributeValueSuggestions(scanner.getTokenEnd());
                        case wxmlLanguageTypes_1.ScannerState.AfterOpeningEndTag:
                            return collectCloseTagSuggestions(scanner.getTokenOffset() - 1, false);
                        case wxmlLanguageTypes_1.ScannerState.WithinContent:
                            return collectCharacterEntityProposals();
                    }
                }
                break;
            case wxmlLanguageTypes_1.TokenType.EndTagOpen:
                if (offset <= scanner.getTokenEnd()) {
                    let afterOpenBracket = scanner.getTokenOffset() + 1;
                    let endOffset = scanNextForEndPos(wxmlLanguageTypes_1.TokenType.EndTag);
                    return collectCloseTagSuggestions(afterOpenBracket, false, endOffset);
                }
                break;
            case wxmlLanguageTypes_1.TokenType.EndTag:
                if (offset <= scanner.getTokenEnd()) {
                    let start = scanner.getTokenOffset() - 1;
                    while (start >= 0) {
                        let ch = text.charAt(start);
                        if (ch === '/') {
                            return collectCloseTagSuggestions(start, false, scanner.getTokenEnd());
                        }
                        else if (!isWhiteSpace(ch)) {
                            break;
                        }
                        start--;
                    }
                }
                break;
            case wxmlLanguageTypes_1.TokenType.StartTagClose:
                if (offset <= scanner.getTokenEnd()) {
                    if (currentTag) {
                        return collectAutoCloseTagSuggestion(scanner.getTokenEnd(), currentTag);
                    }
                }
                break;
            case wxmlLanguageTypes_1.TokenType.Content:
                if (offset <= scanner.getTokenEnd()) {
                    return collectCharacterEntityProposals();
                }
                break;
            default:
                if (offset <= scanner.getTokenEnd()) {
                    return result;
                }
                break;
        }
        token = scanner.scan();
    }
    return result;
}
exports.default = doComplete;
function isQuote(s) {
    return /^["']*$/.test(s);
}
function isWhiteSpace(s) {
    return /^\s*$/.test(s);
}
function isFollowedBy(s, offset, intialState, expectedToken) {
    let scanner = wxmlScanner_1.createScanner(s, offset, intialState);
    let token = scanner.scan();
    while (token === wxmlLanguageTypes_1.TokenType.Whitespace) {
        token = scanner.scan();
    }
    return token === expectedToken;
}
function getWordStart(s, offset, limit) {
    while (offset > limit && !isWhiteSpace(s[offset - 1])) {
        offset--;
    }
    return offset;
}
function getWordEnd(s, offset, limit) {
    while (offset < limit && !isWhiteSpace(s[offset])) {
        offset++;
    }
    return offset;
}
//# sourceMappingURL=wxmlComplete.js.map